c*************************************************************************
c IO_WRITE_PROPER
c*************************************************************************
c accumulate array of orbital elements for test particles and planets,
c filter it (by the given method) and write output to a real*4 binary file.
c
c Input:
c  t         ==> current time (real scalar)
c  nbod      ==> number of massive bodies (int scalar)
c  ntp       ==> number of test particles (int scalar)
c  istat     ==>  status of the test paricles
c  oname     ==> output file name (character string) 
c  iu        ==> unit number to write to
c  fopenstat ==> the status flag for the open statements
c
c other parameters or data in common blocks:
c  /proper/
c  /elements/
c  /elements_mean/
c
c Output:
c  stored in internal arrays and then written to binary file `oname'
c
c Authors: Miroslav Broz, miroslav.broz@email.cz
c Date:    Apr 23rd 2003

      subroutine io_write_proper(t,nbod,ntp,istat,oname,iu,fopenstat)

      include 'swift.inc'
      include 'filter.inc'
      include 'flt_cb.inc'
      include 'proper.inc'
      include 'oscel_cb.inc'
      include 'meanel_cb.inc'

      integer nbod, ntp, iu
      integer istat(NTPMAX,NSTAT)
      real*8 t
      character*(*) oname, fopenstat

c  large buffer for averaging (dimensions are: width of the
c  window, number of elements and PLs+TPs);
c  the same definition should be also in prop_shift subroutine
      real*8 prop_elmts(PMAXN,7,1:PMAXNTP)
      real*8 prop_time(PMAXN)
      save prop_elmts, prop_time

c  temporal variables
      integer i,j,k,i1st,i2nd,PN,ierr,id
      real*8 capa,e,inc,peri,node,varpi,lambda,sigma,Omega,
     :  Omega_PL,varpi_PL,lambda_PL
      real*8 tstart,tstop,tout,eps
      save i1st,i2nd,PN,tstart,tstop
      real*8 data_sep,minfreq,maxfreq
      save data_sep,minfreq,maxfreq
      real*8 input(PMAXN,2),amp,freq
      real*8 g(NTPMAX), s(NTPMAX)
      real*8 sigmalast(NTPMAX)
      save sigmalast

c  functions
      real*8 arr_avg, arr_minmax, zero2pi
      integer arr_minind, arr_maxind, ang_sgndot
      logical ang_eq

c  data
      data i1st,i2nd,PN,eps/0,0,1,1.d-3/

c=======================================================================
c  ... executable code

c  return immediately, if the filter is switched off or
c  mean or osculating elements were NOT updated!
      if ((prop_ftype.eq.0).or.
     :  (.not.((upflg_mean).or.
     :  ((upflg_osc).and.(prop_ftype.eq.4)))))
     :  then
        return
      endif
      upflg_mean = .false.
      upflg_osc = .false.

c  first time through
      if (i1st.eq.0) then
        if (i2nd.eq.0) then

c  try to open outfile (right at the beginning of the integration)
          call io_open(iu,oname,fopenstat,'UNFORMATTED',ierr)
          if (ierr.ne.0) then
            write(*,*) ' SWIFT ERROR: in io_write_proper: '
            write(*,*) '     Could not open binary output file:'
            call util_exit(1)
          endif
          close(iu)

c  initial values
          tstart = t
          tstop = tstart + prop_win
          do i = 1, ntp
            sigmalast(i) = 0.d0
          enddo
          i2nd = 1

c  2nd time through
        else

c  init for FMFT filter (determine time-step)
          if (FNdtout.gt.0) then
            data_sep = FNdtout*dtfilter
          else
            data_sep = (t - tstart)
          endif
          data_sep = data_sep/365.25d0

          minfreq = fmft_minf /180.d0/3600.d0*pi * data_sep
          maxfreq = fmft_maxf /180.d0/3600.d0*pi * data_sep

          i1st = 1
        endif
      endif
c
c  possibly skip some amount of data (ie. if twin.lt.dt)
c
      if (t.lt.tstart-eps) then
        return
      endif
c
c  decide on filter type
c
      if (prop_ftype.eq.1) then

c=======================================================================
c
c  running average filter
c  (it produces exactly the same results as off-line version mean2prop)
c
        prop_time(PN) = t
        do i = 1, ntp
          if (istat(i,1).eq.0) then
            prop_elmts(PN,1,i) = mean_elmts(1,i)
            prop_elmts(PN,2,i) = mean_elmts(6,i)
            prop_elmts(PN,3,i) = mean_elmts(7,i)
          endif
        enddo

c  the buffer is filled-up => write the output
        if (t.gt.tstop) then
          tout = t - 0.5d0*prop_win - 0.5d0*FNinibuf*dtfilter
          call io_open_proper(iu,oname,fopenstat,'UNFORMATTED',ierr)
          if (prop_writer8) then
            call io_write_hdr_r8(iu,tout,0,ntp,istat)
          else
            call io_write_hdr_r(iu,tout,0,ntp,istat)
          endif
          do i = 1, ntp
            if (istat(i,1).eq.0) then
              capa = arr_avg(prop_elmts(1,1,i), PN)
              e = arr_avg(prop_elmts(1,2,i), PN)
              inc = arr_avg(prop_elmts(1,3,i), PN)
              if (prop_writer8) then
                call io_write_line_r8(iu,i,capa,e,inc,0.0d0,0.0d0,0.0d0)
              else
                call io_write_line_r(iu,i,capa,e,inc,0.0d0,0.0d0,0.0d0)
              endif
            endif
          enddo
          close(iu)

c  shift the prop_elmts buffer back in time
          call prop_shift(tstart,tstop,PN,ntp,prop_time,prop_elmts,3)
        endif

        PN = PN + 1

      elseif (prop_ftype.eq.2) then

c=======================================================================
c
c  running minimum (or maximum) filter
c
        prop_time(PN) = t
        do i = 1, ntp
          if (istat(i,1).eq.0) then
            prop_elmts(PN,1,i) = mean_elmts(1,i)
            prop_elmts(PN,2,i) = mean_elmts(6,i)
            prop_elmts(PN,3,i) = mean_elmts(7,i)
            varpi = atan2(mean_elmts(3,i), mean_elmts(2,i))
            node = atan2(mean_elmts(5,i), mean_elmts(4,i))
            peri = zero2pi(varpi-node)
            prop_elmts(PN,4,i) = node
            prop_elmts(PN,5,i) = peri
          endif
        enddo

c  the buffer is filled-up => write the output
        if (t.gt.tstop) then
          tout = t - 0.5d0*prop_win - 0.5d0*FNinibuf*dtfilter
          call io_open_proper(iu,oname,fopenstat,'UNFORMATTED',ierr)
          if (prop_writer8) then
            call io_write_hdr_r8(iu,tout,0,ntp,istat)
          else
            call io_write_hdr_r(iu,tout,0,ntp,istat)
          endif
c  decide whether use minimum or maxium of the orbital element no. prop_ielmt
          if (prop_minmax) then
            do i = 1, ntp
              if (istat(i,1).eq.0) then
                j = arr_minind(prop_elmts(1,prop_ielmt,i), PN)
                capa = prop_elmts(j, 1, i)
                e    = prop_elmts(j, 2, i)
                inc  = prop_elmts(j, 3, i)
                node = prop_elmts(j, 4, i)
                peri = prop_elmts(j, 5, i)
                if (prop_writer8) then
                  call io_write_line_r8(iu,i,capa,e,inc,peri,node,0.0d0)
                else
                  call io_write_line_r(iu,i,capa,e,inc,peri,node,0.0d0)
                endif
              endif
            enddo
c  maximum
          else
            do i = 1, ntp
              if (istat(i,1).eq.0) then
                j = arr_maxind(prop_elmts(1,prop_ielmt,i), PN)
                capa = prop_elmts(j, 1, i)
                e    = prop_elmts(j, 2, i)
                inc  = prop_elmts(j, 3, i)
                node = prop_elmts(j, 4, i)
                peri = prop_elmts(j, 5, i)
                if (prop_writer8) then
                  call io_write_line_r8(iu,i,capa,e,inc,peri,node,0.0d0)
                else
                  call io_write_line_r(iu,i,capa,e,inc,peri,node,0.0d0)
                endif
              endif
            enddo
          endif
          close(iu)

c  shift the prop_elmts buffer back in time
          call prop_shift(tstart,tstop,PN,ntp,prop_time,prop_elmts,5)
        endif

        PN = PN + 1

      elseif (prop_ftype.eq.3) then

c=======================================================================
c
c  running minimum (or maximum) filter for individual elements
c
        prop_time(PN) = t
        do i = 1, ntp
          if (istat(i,1).eq.0) then
            prop_elmts(PN,1,i) = mean_elmts(1,i)
            prop_elmts(PN,2,i) = mean_elmts(6,i)
            prop_elmts(PN,3,i) = mean_elmts(7,i)
          endif
        enddo

c  the buffer is filled-up => write the output
        if (t.gt.tstop) then
          tout = t - 0.5d0*prop_win - 0.5d0*FNinibuf*dtfilter
          call io_open_proper(iu,oname,fopenstat,'UNFORMATTED',ierr)
          if (prop_writer8) then
            call io_write_hdr_r8(iu,tout,0,ntp,istat)
          else
            call io_write_hdr_r(iu,tout,0,ntp,istat)
          endif
c  minimum or maxium of the orbital elements
          do i = 1, ntp
            if (istat(i,1).eq.0) then
              capa = arr_minmax(prop_elmts(1,1,i), PN, prop_mmel(1))
              e    = arr_minmax(prop_elmts(1,2,i), PN, prop_mmel(2))
              inc  = arr_minmax(prop_elmts(1,3,i), PN, prop_mmel(3))
              if (prop_writer8) then
                call io_write_line_r8(iu,i,capa,e,inc,0.0d0,0.0d0,0.0d0)
              else
                call io_write_line_r(iu,i,capa,e,inc,0.0d0,0.0d0,0.0d0)
              endif
            endif
          enddo
          close(iu)

c  shift the prop_elmts buffer back in time
          call prop_shift(tstart,tstop,PN,ntp,prop_time,prop_elmts,3)
        endif

        PN = PN + 1

      elseif (prop_ftype.eq.4) then

c=======================================================================
c
c  sigma & varpi-varpi_PL \simeq 0 filter (resonant elements)
c  critical angle of a resonance: sigma = (p+q)/q*lambda_PL - p/q*lambda_TP - varpi_TP
c
c  we actually do NOT need the buffer, we must only check the above condition
c  and take THE LAST OSCULATING elements, instead of mean one! (lambda and
c  sigma changes very quickly and we need a sufficient sampling)

        do i = 1, ntp
          if (istat(i, 1).eq.0) then
            Omega = osc_elmts(4,i)
            varpi = zero2pi(Omega + osc_elmts(5,i))
            lambda = zero2pi(varpi + osc_elmts(6,i))
            Omega_PL = osc_elmts(4,prop_plid)
            varpi_PL = zero2pi(Omega_PL + osc_elmts(5,prop_plid))
            lambda_PL = zero2pi(varpi_PL + osc_elmts(6,prop_plid))
            sigma = zero2pi(dble(prop_p+prop_q)/prop_q * lambda_PL
     :        - dble(prop_p)/prop_q * lambda
     :        - varpi)

c  if the condition is true, save the last osculating elements
            if (ang_eq(sigma, prop_sigma, prop_dsigma)
     :        .and.((prop_sigmadot.eq.0).or.
     :        (ang_sgndot(sigmalast(i), sigma).eq.prop_sigmadot))
     :        .and.((prop_dvarpi.lt.0.d0).or.
     :        ang_eq(zero2pi(varpi-varpi_PL),prop_varpi,prop_dvarpi))
     :        .and.((prop_dOmega.lt.0.d0).or.
     :        ang_eq(zero2pi(Omega-Omega_PL),prop_Omega,prop_dOmega)))
     :        then

c  save the OSCULATING elements (regardless short-periodic escilations!)
c  we use the 1st position ONLY in prop_elmts buffer
              do j = 1, 6
                prop_elmts(1,j,i) = osc_elmts(j,i)
              enddo

c  debug code
c              write(*,'(i5,1x,7(1x,f22.10),2(1x,i3))')
c     :          i,
c     :          t/365.25d6,
c     :          osc_elmts(1,i),
c     :          osc_elmts(2,i),
c     :          sigma*degrad,
c     :          zero2pi(varpi-varpi_PL)*degrad,
c     :          zero2pi(Omega-Omega_PL)*degrad,
c     :          sigmalast(i)*degrad,
c     :          ang_sgndot(sigmalast(i), sigma),
c     :          prop_sigmadot
            endif

c  save last sigma due to d sigma/d t monitoring
            sigmalast(i) = sigma
          endif
        enddo

c  the buffer is filled-up => write the output
        if (t.gt.tstop) then
          tout = t - 0.5d0*prop_win
          call io_open_proper(iu,oname,fopenstat,'UNFORMATTED',ierr)
          if (prop_writer8) then
            call io_write_hdr_r8(iu,tout,0,ntp,istat)
          else
            call io_write_hdr_r(iu,tout,0,ntp,istat)
          endif
c  minimum or maxium of the orbital elements
          do i = 1, ntp
            if (istat(i,1).eq.0) then
              if (prop_writer8) then
                call io_write_line_r8(iu,i,
     :            prop_elmts(1,1,i),
     :            prop_elmts(1,2,i),
     :            prop_elmts(1,3,i),
     :            prop_elmts(1,4,i),
     :            prop_elmts(1,5,i),
     :            prop_elmts(1,6,i))
              else
                call io_write_line_r(iu,i,
     :            prop_elmts(1,1,i),
     :            prop_elmts(1,2,i),
     :            prop_elmts(1,3,i),
     :            prop_elmts(1,4,i),
     :            prop_elmts(1,5,i),
     :            prop_elmts(1,6,i))
              endif
            endif
          enddo
          close(iu)
          tstop = tstop + prop_dt
          tstart = tstop - prop_win
        endif

      elseif (prop_ftype.eq.5) then

c=======================================================================
c
c  FMFT filter, calls subroutine fmft.c (in C) by D. Nesvorny
c
        prop_time(PN) = t
        do i = 1, ntp
          if (istat(i, 1).eq.0) then
            prop_elmts(PN,1,i) = mean_elmts(1,i)
            prop_elmts(PN,2,i) = mean_elmts(2,i)
            prop_elmts(PN,3,i) = mean_elmts(3,i)
            prop_elmts(PN,4,i) = mean_elmts(4,i)
            prop_elmts(PN,5,i) = mean_elmts(5,i)
          endif
        enddo

c  the buffer is filled-up (2^N samples!) => write the output
        if (PN.ge.fmft_ndata) then
          tout = t
          call io_open_proper(iu,oname,fopenstat,'UNFORMATTED',ierr)
          if (prop_writer8) then
            call io_write_hdr_r8(iu,tout,0,ntp,istat)
          else
            call io_write_hdr_r(iu,tout,0,ntp,istat)
          endif

c  call FMFT for each TP and for 2 pairs of elements (h and k, p and q)
          do i = 1, ntp
            if (istat(i, 1).eq.0) then

c  a (semimajor axis) is an usual running average
              capa = arr_avg(prop_elmts(1,1,i), PN)

c  p and q first
              do j = 1, fmft_ndata
                input(j, 1) = prop_elmts(j,2,i)
                input(j, 2) = prop_elmts(j,3,i)
              enddo

              call fmft_call(input,minfreq,maxfreq,data_sep,amp,freq)
              e = amp
              g(i) = freq

c  k and h second
              do j = 1, fmft_ndata
                input(j, 1) = prop_elmts(j,4,i)
                input(j, 2) = prop_elmts(j,5,i)
              enddo

              call fmft_call(input,minfreq,maxfreq,data_sep,amp,freq)
              inc = 2.0d0*asin(amp)
              s(i) = freq

c  write output (if istat.eq.0)
              if (prop_writer8) then
                call io_write_line_r8(iu,i,capa,e,inc,0.0d0,0.0d0,0.0d0)
              else
                call io_write_line_r(iu,i,capa,e,inc,0.0d0,0.0d0,0.0d0)
              endif

            endif
          enddo
          close(iu)

c  eventually write frequencies
          if (fmft_writegs) then
            call io_write_fmft(tout,0,ntp,istat,g,s,iu,fopenstat)
          endif

c  shift the prop_elmts buffer back in time
          call prop_shift(tstart,tstop,PN,ntp,prop_time,prop_elmts,3)
        endif

        PN = PN + 1

      endif

      return
      end

c****************************************************************
      real*8 function arr_avg(a, n)
c****************************************************************
      implicit none
      integer n
      real*8 a(n)

      integer i
      real*8 suma

      suma = 0.d0
      do i = 1, n
        suma = suma + a(i)
      enddo
      arr_avg = suma / n

      return
      end

c****************************************************************
      integer function arr_minind(a, n)
c****************************************************************
      implicit none
      integer n
      real*8 a(n)

      integer i, mini
      real*8 mina

      mina = a(1)
      mini = 1
      do i = 2, n
        if (a(i).lt.mina) then
          mina = a(i)
          mini = i
        endif
      enddo
      arr_minind = mini

      return
      end

c****************************************************************
      integer function arr_maxind(a, n)
c****************************************************************
      implicit none
      integer n
      real*8 a(n)

      integer i, maxi
      real*8 maxa

      maxa = a(1)
      maxi = 1
      do i = 2, n
        if (a(i).gt.maxa) then
          maxa = a(i)
          maxi = i
        endif
      enddo
      arr_maxind = maxi

      return
      end

c****************************************************************
      real*8 function arr_minmax(a, n, minmax)
c****************************************************************
      implicit none
      integer n
      real*8 a(n)
      logical minmax

      integer i
c  functions
      integer arr_minind, arr_maxind

      if (minmax) then
        i = arr_minind(a, n)
      else
        i = arr_maxind(a, n)
      endif
      arr_minmax = a(i)

      return
      end

c****************************************************************
      subroutine arr_shift(a, n, k)
c****************************************************************
      implicit none
      integer n, k
      real*8 a(n)

      integer i

      do i = k+1, n
        a(i-k) = a(i)
      enddo

      return
      end

c****************************************************************
      subroutine prop_shift(tstart,tstop,PN,ntp,
     :  prop_time,prop_elmts,nelmts)
c****************************************************************
c
c  shift the prop_* buffers back in time
c
      implicit none
      include 'filter.inc'
      include 'proper.inc'

      integer PN,ntp,nelmts
      real*8 tstart,tstop
      real*8 prop_time(PMAXN)
      real*8 prop_elmts(PMAXN,7,1:PMAXNTP)

      integer i,j,k,id

      tstop = tstop + prop_dt
      tstart = tstop - prop_win
      if (tstart.ge.prop_time(PN)) then
        PN = 0
      else
        k = 1
        do while (prop_time(k).lt.tstart)
          k = k + 1
        enddo
        call arr_shift(prop_time, PN, k)
        do i = 1, ntp
          id = i
          do j = 1, nelmts
            call arr_shift(prop_elmts(1,j,id), PN, k)
          enddo 
        enddo
        PN = PN - k
      endif

      return
      end

c****************************************************************
      subroutine fmft_call(input,minfreq,maxfreq,data_sep,amp,freq)
c****************************************************************
c
c  call FMFT subroutine, discard small or known frequencies
c  and return a proper element
c
      include 'swift.inc'
      include 'filter.inc'
      include 'proper.inc'

      real*8 input(MAX_FMFT_NDATA, 2)
      real*8 output(MAX_FMFT_NFREQ, 3*MAX_FMFT_FLAG)
      real*8 minfreq, maxfreq, data_sep
      real*8 freq, amp, phase

      integer j, k, err
      logical fmft_print
c  functions
      integer fmft

      err = fmft(output, fmft_nfreq, minfreq, maxfreq,
     :  fmft_flag, input, fmft_ndata)

c  check the error flag, but do NOT stop
      if (err.eq.0) then
        write(*,20)
20      format('Unknown error in fmft C-function.')
c        stop
      endif

      do j = 2, fmft_nfreq+1
        freq = output(j, 3*fmft_flag-1)
     :    *((180.d0/pi)*3600.d0)/data_sep
      enddo

      j = 2
      fmft_print = .false.
      do while ((j.le.fmft_nfreq+1).and.(.not.fmft_print))
        freq = output(j, 3*fmft_flag-1)
     :    *((180.d0/pi)*3600.d0)/data_sep

        fmft_print = .true.
c  discard too small frequency
        if (abs(freq).lt.fmft_freqsmall) then
          fmft_print = .false.
        else
c  discard well known planetary frequences
          do k = 1, fmft_nfreqknown
            if (abs(freq - fmft_freqknown(k)).lt.fmft_freqtoler) then
              fmft_print = .false.
            endif
          enddo
        endif

        if (fmft_print) then
          amp = output(j, 3*fmft_flag)
c          phase = output(i, 3*fmft_flag+1) *180.d0/pi
c          if (phase.lt.0) phase = phase + 360.d0
        endif

        j = j + 1
      enddo

      return
      end

c****************************************************************
      logical function ang_eq(a, b, toler)
c****************************************************************
c
c  Test, if two angles <0, 2pi) are equal within tolerance.
c
      real*8 a, b, toler

      real*8 PI,TWOPI
      parameter (PI = 3.14159265358979D0)
      parameter (TWOPI = 2.0D0 * PI)

      if ((abs(a-b).lt.toler).or.
     :  (abs(a-b-TWOPI).lt.toler).or.
     :  (abs(a-b+TWOPI).lt.toler)) then
        ang_eq = .true.
      else
        ang_eq = .false.
      endif

      return
      end

c****************************************************************
      integer function ang_sgndot(a, b)
c****************************************************************
c
c  Signum of d angle/d t from two consecutive values a, b,
c  asure angles near 0 and 2pi.
c
      real*8 a, b

      real*8 PI,TWOPI
      parameter (PI = 3.14159265358979D0)
      parameter (TWOPI = 2.0D0 * PI)

      integer isgn

      if (abs(b-a-TWOPI).lt.abs(b-a)) then
        ang_sgndot = isgn(b-a-TWOPI)
      else if (abs(b-a+TWOPI).lt.abs(b-a)) then
        ang_sgndot = isgn(b-a+TWOPI)
      else
        ang_sgndot = isgn(b-a)
      endif

      return
      end

c****************************************************************
      integer function isgn(x)
c****************************************************************
c
c  Return an integer signum of a double precision argument.
c
      real*8 x
 
      if (x.gt.0.d0) then
        isgn = 1
      elseif (x.lt.0.d0) then
        isgn = -1
      else
        isgn = 0
      endif

      return
      end

c****************************************************************

